"""
Production RWKV.
RWKV - Receptance Weighted Key Value 
This project aims to make RWKV Accessible to everyone using a Hugging Face like OOP interface, while staying close to the research.
"""

__version__ = "0.0.1"
__author__ = 'Michael Chung (ArEnSc)'
__credits__ = 'Blink DL, EmElleE'